package rition.service.collector;

import com.github.yitter.idgen.YitIdHelper;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.data.redis.core.script.RedisScript;
import org.springframework.kafka.core.KafkaTemplate;
import org.springframework.stereotype.Service;
import rition.common.data.dao.mapper.MetricRecordMapper;
import rition.common.data.dto.service.MetricDataDto;
import rition.common.data.entity.MetricRecordEntity;
import rition.common.data.enums.Constants;
import rition.service.collector.configure.CollectorServiceKafkaConfigure;

import java.time.Instant;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * DataCollectorService 指标数据收集类，负责收集来自云主机探针上报的数据，并将其入库等
 */
@Slf4j
@Service
public class MetricCollectingService {

    private final KafkaTemplate<String, MetricDataDto> kafkaTemplate;

    private final String collectedDataTopic;
    private final MetricRecordMapper metricRecordMapper;
    private final RedisTemplate<String, MetricDataDto> redisTemplate;

    private final RedisScript<Object> constLenQueueRedisScript;

    public MetricCollectingService(KafkaTemplate<String, MetricDataDto> kafkaTemplate,
                                   CollectorServiceKafkaConfigure collectorServiceKafkaConfigure,
                                   MetricRecordMapper metricRecordMapper,
                                   RedisTemplate<String, MetricDataDto> redisTemplate,

                                   @Qualifier("constLenQueueRedisScript") RedisScript<Object> constLenQueueRedisScript) {

        this.kafkaTemplate = kafkaTemplate;
        this.collectedDataTopic = collectorServiceKafkaConfigure.getDataCollecting().getTopic();
        this.metricRecordMapper = metricRecordMapper;
        this.redisTemplate = redisTemplate;

        this.constLenQueueRedisScript = constLenQueueRedisScript;
    }

    private static final int METRIC_NUMS = 22;

    // 目前设定为50组数据作为一个批次
    private static final int DEFAULT_DB_BATCH_SIZE = 50;

    /**
     * 接收处理好的监控指标数据，以时间戳为分组入库和发送mq
     *
     * @param collectedMetricData 监控指标数据，以时间戳为组
     */
    public void receiveData(Map<Long, Map<String, MetricDataDto>> collectedMetricData) {
        List<MetricRecordEntity> batchSaveEntityList = new ArrayList<>(DEFAULT_DB_BATCH_SIZE);
        int turn = 0;
        for (Long timestamp : collectedMetricData.keySet()) {
            Map<String, MetricDataDto> instanceMetricData = collectedMetricData.get(timestamp);
            for (String instanceId : instanceMetricData.keySet()) {
                MetricDataDto metricDataDto = instanceMetricData.get(instanceId);

                MetricRecordEntity entity = new MetricRecordEntity();
                entity.setId(YitIdHelper.nextId());
                entity.setInstanceId(instanceId);
                entity.setMetricData(metricDataDto.getData());
                var time = Instant.ofEpochSecond(metricDataDto.getTimestamp());
                entity.setTime(time);
                entity.setUpdateTime(time);
                entity.setStatus(0);

                // kafka发布
                kafkaTemplate.send(this.collectedDataTopic, metricDataDto);

                // 分批入库，只是针对大批量请求的处理
                batchSaveEntityList.add(entity);
                if (turn >= DEFAULT_DB_BATCH_SIZE) {
                    metricRecordMapper.insertBatchSomeColumn(batchSaveEntityList);
                    // 下一轮
                    batchSaveEntityList.clear();
                    turn = 0;
                } else {
                    turn++;
                }

                // 缓存最近数据
                redisTemplate.execute(
                        constLenQueueRedisScript,
                        List.of(Constants.RedisKeys.RECENT_METRIC_CACHE),
                        Constants.MAX_METRIC_CACHE_NUM, metricDataDto
                );
            }
        }

        // 还有没入库的
        if (!batchSaveEntityList.isEmpty()) {
            metricRecordMapper.insertBatchSomeColumn(batchSaveEntityList);
        }
    }
}
